from __future__ import annotations

import contextlib
import hashlib
import os
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional
from urllib.parse import urlparse

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData, FileDataSourceMetadata
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from fsspec.implementations.sftp import SFTPFileSystem

CONNECTOR_TYPE = "sftp"


def _strip_leading_slash(path: str) -> str:
    """Strip one leading slash, preserving double-slash absolute path indicators.
    Example: sftp://host//home → /home (absolute), sftp://host/data → data (relative)."""
    return path[1:] if path.startswith("/") else path


# Patch hashlib.md5 for FIPS-enabled OpenSSL (common in Kubernetes).
# Paramiko uses MD5 solely for logging human-readable host key fingerprints,
# not for any cryptographic purpose (SSH security uses Ed25519/SHA-256).
# This flag tells OpenSSL the MD5 call is non-cryptographic, which is safe.
_original_md5 = hashlib.md5


def _fips_safe_md5(data=b"", **kwargs):
    kwargs.setdefault("usedforsecurity", False)
    return _original_md5(data, **kwargs)


hashlib.md5 = _fips_safe_md5


class SftpIndexerConfig(FsspecIndexerConfig):
    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        _, ext = os.path.splitext(self.remote_url)
        parsed_url = urlparse(self.remote_url)
        if ext:
            parent_path = Path(parsed_url.path).parent.as_posix()
            self.path_without_protocol = _strip_leading_slash(parent_path)
        else:
            self.path_without_protocol = _strip_leading_slash(parsed_url.path)


class SftpAccessConfig(FsspecAccessConfig):
    password: str = Field(description="Password for sftp connection")


class SftpConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = Field(default_factory=lambda: ["sftp"], init=False)
    access_config: Secret[SftpAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    username: str = Field(description="Username for sftp connection")
    host: Optional[str] = Field(default=None, description="Hostname for sftp connection")
    port: int = Field(default=22, description="Port for sftp connection")
    look_for_keys: bool = Field(
        default=False, description="Whether to search for private key files in ~/.ssh/"
    )
    allow_agent: bool = Field(default=False, description="Whether to connect to the SSH agent.")

    def get_access_config(self) -> dict[str, Any]:
        access_config = {
            "username": self.username,
            "host": self.host,
            "port": self.port,
            "look_for_keys": self.look_for_keys,
            "allow_agent": self.allow_agent,
            "password": self.access_config.get_secret_value().password,
        }
        return access_config

    @contextmanager
    @requires_dependencies(["paramiko", "fsspec"], extras="sftp")
    def get_client(self, protocol: str) -> Generator["SFTPFileSystem", None, None]:
        # The paramiko.SSHClient() client that's opened by the SFTPFileSystem
        # never gets closed so explicitly adding that as part of this context manager.
        # skip_instance_cache=True is required to prevent fsspec from returning a cached
        # instance whose SSH connection was closed by a previous context manager exit.
        from fsspec import get_filesystem_class

        client: SFTPFileSystem = get_filesystem_class(protocol)(
            skip_instance_cache=True,
            **self.get_access_config(),
        )
        yield client
        client.client.close()


@dataclass
class SftpIndexer(FsspecIndexer):
    connection_config: SftpConnectionConfig
    index_config: SftpIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def __post_init__(self):
        super().__post_init__()
        parsed_url = urlparse(self.index_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port

    def precheck(self) -> None:
        self.log_operation_start(
            "Connection validation",
            protocol=self.index_config.protocol,
            path=self.index_config.path_without_protocol,
        )

        try:
            with self.connection_config.get_client(protocol=self.index_config.protocol) as client:
                files = client.ls(path=self.index_config.path_without_protocol, detail=True)
                valid_files = [x.get("name") for x in files if x.get("type") == "file"]
                if not valid_files:
                    self.log_operation_complete("Connection validation", count=0)
                    return
                file_to_sample = valid_files[0]
                self.log_debug(f"attempting to make HEAD request for file: {file_to_sample}")
                client.head(path=file_to_sample)

            self.log_connection_validated(
                connector_type=self.connector_type,
                endpoint=f"{self.index_config.protocol}://{self.index_config.path_without_protocol}",
            )

        except Exception as e:
            self.log_connection_failed(
                connector_type=self.connector_type,
                error=e,
                endpoint=f"{self.index_config.protocol}://{self.index_config.path_without_protocol}",
            )
            raise self.wrap_error(e=e)

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["name"]
        date_created = str(file_info.get("time").timestamp()) if "time" in file_info else None
        date_modified = str(file_info.get("mtime").timestamp()) if "mtime" in file_info else None

        file_size = file_info.get("size") if "size" in file_info else None

        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
        }
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class SftpDownloaderConfig(FsspecDownloaderConfig):
    remote_url: str = Field(description="Remote fsspec URL formatted as `protocol://dir/path`")


@dataclass
class SftpDownloader(FsspecDownloader):
    protocol: str = "sftp"
    connection_config: SftpConnectionConfig
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    download_config: Optional[SftpDownloaderConfig] = field(default_factory=SftpDownloaderConfig)

    def __post_init__(self):
        super().__post_init__()
        parsed_url = urlparse(self.download_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port


class SftpUploaderConfig(FsspecUploaderConfig):
    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        parsed_url = urlparse(self.remote_url)
        self.path_without_protocol = _strip_leading_slash(parsed_url.path)


@dataclass
class SftpUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: SftpConnectionConfig
    upload_config: SftpUploaderConfig = field(default=None)

    def __post_init__(self):
        super().__post_init__()
        parsed_url = urlparse(self.upload_config.remote_url)
        self.connection_config.host = parsed_url.hostname or self.connection_config.host
        self.connection_config.port = parsed_url.port or self.connection_config.port

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        path_str = str(path.resolve())
        upload_path = self.get_upload_path(file_data=file_data)
        self.log_upload_start(file_path=path_str, destination=upload_path.as_posix())
        try:
            with self.connection_config.get_client(protocol=self.upload_config.protocol) as client:
                # fsspec's SFTPFileSystem.put() bypasses the base class's mkdirs
                # (going straight to paramiko), so we create parent dirs explicitly.
                client.makedirs(upload_path.parent.as_posix(), exist_ok=True)
                client.upload(lpath=path_str, rpath=upload_path.as_posix())
        except Exception as e:
            self.log_error(
                "File upload failed",
                error=e,
                context={"file_path": path_str, "destination": upload_path.as_posix()},
            )
            raise self.wrap_error(e=e)
        self.log_upload_complete(file_path=path_str, destination=upload_path.as_posix())

    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        self.run(path=path, file_data=file_data, **kwargs)

    def precheck(self) -> None:
        self.log_operation_start("Connection validation", protocol=self.upload_config.protocol)

        try:
            with self.connection_config.get_client(protocol=self.upload_config.protocol) as client:
                upload_path = Path(self.upload_config.path_without_protocol) / "_empty"
                # Create parent directories if they don't exist
                client.makedirs(upload_path.parent.as_posix(), exist_ok=True)
                client.write_bytes(path=upload_path.as_posix(), value=b"")
                # Best-effort cleanup - don't fail if user lacks delete permissions
                with contextlib.suppress(Exception):
                    client.rm(path=upload_path.as_posix())

            self.log_connection_validated(
                connector_type=self.connector_type,
                endpoint=f"{self.upload_config.protocol}://{self.upload_config.path_without_protocol}",
            )

        except Exception as e:
            self.log_connection_failed(
                connector_type=self.connector_type,
                error=e,
                endpoint=f"{self.upload_config.protocol}://{self.upload_config.path_without_protocol}",
            )
            raise self.wrap_error(e=e)


sftp_source_entry = SourceRegistryEntry(
    indexer=SftpIndexer,
    indexer_config=SftpIndexerConfig,
    downloader=SftpDownloader,
    downloader_config=SftpDownloaderConfig,
    connection_config=SftpConnectionConfig,
)

sftp_destination_entry = DestinationRegistryEntry(
    uploader=SftpUploader,
    uploader_config=SftpUploaderConfig,
    connection_config=SftpConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



import itertools
import json
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Iterable, Optional, Sequence, TypeVar, cast
from uuid import NAMESPACE_DNS, uuid5

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.logger import logger
from unstructured_ingest.utils import ndjson
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from pandas import DataFrame

DATE_FORMATS = ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d+%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z")

T = TypeVar("T")
IterableT = Iterable[T]


def split_dataframe(df: "DataFrame", chunk_size: int = 100) -> Generator["DataFrame", None, None]:
    num_chunks = len(df) // chunk_size + 1
    for i in range(num_chunks):
        yield df[i * chunk_size : (i + 1) * chunk_size]


def batch_generator(iterable: IterableT, batch_size: int = 100) -> IterableT:
    """A helper function to break an iterable into batches of size batch_size."""
    it = iter(iterable)
    chunk = tuple(itertools.islice(it, batch_size))
    while chunk:
        yield chunk
        chunk = tuple(itertools.islice(it, batch_size))


def generator_batching_wbytes(
    iterable: IterableT,
    batch_size_limit_bytes: Optional[int] = None,
    max_batch_size: Optional[int] = None,
) -> IterableT:
    if not batch_size_limit_bytes and not max_batch_size:
        return iterable
    """A helper function to break an iterable into chunks of specified bytes."""
    current_batch, current_batch_size = [], 0

    for item in iterable:
        item_size_bytes = len(json.dumps(item).encode("utf-8"))
        if batch_size_limit_bytes and current_batch_size + item_size_bytes > batch_size_limit_bytes:
            yield current_batch
            current_batch, current_batch_size = [item], item_size_bytes
            continue
        if max_batch_size and len(current_batch) + 1 > max_batch_size:
            yield current_batch
            current_batch, current_batch_size = [item], item_size_bytes
            continue

        current_batch.append(item)
        current_batch_size += item_size_bytes

    if current_batch:
        yield current_batch


def flatten_dict(
    dictionary: dict[str, Any],
    parent_key: str = "",
    separator: str = "_",
    flatten_lists: bool = False,
    remove_none: bool = False,
    keys_to_omit: Optional[Sequence[str]] = None,
) -> dict[str, Any]:
    """Flattens a nested dictionary into a single level dictionary.

    keys_to_omit is a list of keys that don't get flattened. If omitting a nested key, format as
    {parent_key}{separator}{key}. If flatten_lists is True, then lists and tuples are flattened as
    well. If remove_none is True, then None keys/values are removed from the flattened
    dictionary.
    """
    keys_to_omit = keys_to_omit if keys_to_omit else []
    flattened_dict: dict[str, Any] = {}
    for key, value in dictionary.items():
        new_key = f"{parent_key}{separator}{key}" if parent_key else key
        if new_key in keys_to_omit:
            flattened_dict[new_key] = value
        elif value is None and remove_none:
            continue
        elif isinstance(value, dict):
            value = cast("dict[str, Any]", value)
            flattened_dict.update(
                flatten_dict(
                    value, new_key, separator, flatten_lists, remove_none, keys_to_omit=keys_to_omit
                ),
            )
        elif isinstance(value, (list, tuple)) and flatten_lists:
            value = cast("list[Any] | tuple[Any]", value)
            for index, item in enumerate(value):
                flattened_dict.update(
                    flatten_dict(
                        {f"{new_key}{separator}{index}": item},
                        "",
                        separator,
                        flatten_lists,
                        remove_none,
                        keys_to_omit=keys_to_omit,
                    )
                )
        else:
            flattened_dict[new_key] = value

    return flattened_dict


def validate_date_args(date: Optional[str] = None) -> bool:
    """Validate whether the provided date string satisfies any of the supported date formats.

    Used by unstructured/ingest/connector/biomed.py

    Returns `True` if the date string satisfies any of the supported formats, otherwise raises
    `ValueError`.

    Supported Date Formats:
        - 'YYYY-MM-DD'
        - 'YYYY-MM-DDTHH:MM:SS'
        - 'YYYY-MM-DD+HH:MM:SS'
        - 'YYYY-MM-DDTHH:MM:SS±HHMM'
    """
    if not date:
        raise ValueError("The argument date is None.")

    for format in DATE_FORMATS:
        try:
            datetime.strptime(date, format)
            return True
        except ValueError:
            pass

    raise ValueError(
        f"The argument {date} does not satisfy the format:"
        f" YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS or YYYY-MM-DD+HH:MM:SS or YYYY-MM-DDTHH:MM:SS±HHMM",
    )


def get_data_by_suffix(path: Path) -> list[dict]:
    with path.open() as f:
        if path.suffix == ".json":
            return json.load(f)
        elif path.suffix == ".ndjson":
            return ndjson.load(f)
        elif path.suffix == ".csv":
            import pandas as pd

            df = pd.read_csv(path)
            return df.to_dict(orient="records")
        elif path.suffix == ".parquet":
            import pandas as pd

            df = pd.read_parquet(path)
            return df.to_dict(orient="records")
        else:
            raise ValueError(f"Unsupported file type: {path}")


def write_data(path: Path, data: list[dict], indent: Optional[int] = 2) -> None:
    with path.open("w") as f:
        if path.suffix == ".json":
            json.dump(data, f, indent=indent, ensure_ascii=False)
        elif path.suffix == ".ndjson":
            ndjson.dump(data, f, ensure_ascii=False)
        else:
            raise IOError("Unsupported file type: {path}")


def get_json_data(path: Path) -> list[dict]:
    with path.open() as f:
        # Attempt by prefix
        if path.suffix == ".json":
            return json.load(f)
        elif path.suffix == ".ndjson":
            return ndjson.load(f)
        try:
            return json.load(f)
        except Exception as e:
            logger.warning(f"failed to read {path} as json: {e}")
        try:
            return ndjson.load(f)
        except Exception as e:
            logger.warning(f"failed to read {path} as ndjson: {e}")
    raise ValueError(f"Unsupported json file: {path}")


@requires_dependencies(["pandas"])
def get_data_df(path: Path) -> "DataFrame":
    import pandas as pd

    with path.open() as f:
        if path.suffix == ".json":
            data = json.load(f)
            return pd.DataFrame(data=data)
        elif path.suffix == ".ndjson":
            data = ndjson.load(f)
            return pd.DataFrame(data=data)
        elif path.suffix == ".csv":
            df = pd.read_csv(path)
            return df
        elif path.suffix == ".parquet":
            df = pd.read_parquet(path)
            return df
        else:
            raise ValueError(f"Unsupported file type: {path}")


def get_enhanced_element_id(element_dict: dict, file_data: FileData) -> str:
    element_id = element_dict.get("element_id")
    new_data = f"{element_id}{file_data.identifier}"
    return str(uuid5(NAMESPACE_DNS, new_data))
